package ch.nostromo.tiffanys.clients.desktop.controller;

import java.awt.Frame;
import java.io.File;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.JOptionPane;
import javax.swing.JTextArea;

import ch.nostromo.lib.controller.NosController;
import ch.nostromo.lib.controller.NosControllerException;
import ch.nostromo.lib.controller.NosResourcesException;
import ch.nostromo.lib.controller.NosSettingsException;
import ch.nostromo.tiffanys.clients.desktop.gui.Desktop;
import ch.nostromo.tiffanys.engines.EngineSettings;
import ch.nostromo.tiffanys.game.GameController;
import ch.nostromo.tiffanys.game.GameControllerException;
import ch.nostromo.tiffanys.game.GameControllerListener;
import ch.nostromo.tiffanys.game.GameControllerResult;
import ch.nostromo.tiffanys.game.board.Board;
import ch.nostromo.tiffanys.game.game.GameException;
import ch.nostromo.tiffanys.game.game.GameState;
import ch.nostromo.tiffanys.game.move.Move;
import ch.nostromo.tiffanys.game.move.MoveInput;
import ch.nostromo.tiffanys.game.pgn.PGN;
import ch.nostromo.tiffanys.game.player.Player;
import ch.nostromo.tiffanys.game.player.PlayerEngine;
import ch.nostromo.tiffanys.game.player.PlayerException;
import ch.nostromo.tiffanys.game.player.PlayerHuman;
import ch.nostromo.tiffanys.game.player.PlayerTools;

public class DesktopController extends NosController implements GameControllerListener {

  private static Logger logger = Logger.getLogger(DesktopController.class.getName());

  private static DesktopController instance = null;

  private final static File desktopSettingsFile = new File(System.getProperty("user.home") + "/.tiffanys", "desktop-" + ch.nostromo.tiffanys.Tiffanys.VERSION + ".xml");

  private DesktopSettings settings = null;

  private GameController currentGameController;

  private Desktop desktop;

  /**
   * Returns the one and only instance of the controller.
   * 
   * @return The instance value
   */
  public static DesktopController getInstance() {
    if (instance == null) {
      instance = new DesktopController();
    }

    return instance;
  }

  /**
   * Initialize the application (read parameters, start logging etc.)
   * 
   * @param args
   *          Command line arguments
   */
  public void init(String[] args) {
    try {

      // Load client settings (if not found, try to save them for the first
      // time)
      try {
        settings = new DesktopSettings(desktopSettingsFile);
      } catch (NosSettingsException err) {
        settings = new DesktopSettings();
        settings.writeSettingsToXml(desktopSettingsFile);
        settings = new DesktopSettings(desktopSettingsFile);
      }

      this.initController(settings);

      // Start initializing
      logger.info("Starting Tiffanys " + ch.nostromo.tiffanys.Tiffanys.VERSION + ".");

      logger.info("Appending Uncaught Exception handling ...");
      Thread.setDefaultUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {

        @SuppressWarnings("synthetic-access")
        public void uncaughtException(Thread t, Throwable e) {
          String msg = "Uncaught Exception catched!";
          logger.log(Level.SEVERE, msg, e);
          showError(msg, e);

        }
      });

      desktop = new Desktop();
      desktop.setVisible(true);

      this.translateComponents(desktop);

      // Create new Game
      launchNewGame(generateDefaultGameController());

   
    } catch (NosSettingsException e) {
      System.err.println(e);
      logger.log(Level.SEVERE, "Unexpected Exception, program halted !", e);
      endApp();
    } catch (NosResourcesException e) {
      System.err.println(e);
      logger.log(Level.SEVERE, "Unexpected Exception, program halted !", e);
      endApp();
    } catch (NosControllerException e) {
      System.err.println(e);
      logger.log(Level.SEVERE, "Unexpected Exception, program halted !", e);
      endApp();
    }

  }

  public GameController generateDefaultGameController() throws NosSettingsException, NosResourcesException {
    // Start new game controller
    Player white = getDefaultWhitePlayer();
    Player black = getDefaultBlackPlayer();

    GameController newGame = new GameController(white, black);
    return newGame;
  }

  public Player getDefaultWhitePlayer() throws NosSettingsException, NosResourcesException {
    if (settings.DEFAULT_GAME_WHITE_ENGINE) {
      return getDefaultPlayerEngine();
    } else {
      return getDefaultPlayerHuman();
    }
  }

  public Player getDefaultBlackPlayer() throws NosResourcesException, NosSettingsException {
    if (settings.DEFAULT_GAME_BLACK_ENGINE) {
      return getDefaultPlayerEngine();
    } else {
      return getDefaultPlayerHuman();
    }
  }

  public PlayerHuman getDefaultPlayerHuman() throws NosResourcesException {
    return PlayerTools.createPlayerHuman(getTranslation("Common.HumanPlayerName"));
  }

  public PlayerEngine getDefaultPlayerEngine() throws NosSettingsException {
    return PlayerTools.createPlayerEngine(settings.DEFAULT_ENGINE_NAME, getDefaultEngineSettings());
  }

  private EngineSettings getDefaultEngineSettings() throws NosSettingsException {
    String fileName = getSettings().DEFAULT_ENGINE_SETTINGS;
    return new EngineSettings(fileName);
  }

  public void launchNewGame(GameController gameController) {
    try {
      if (currentGameController != null) {
        currentGameController.removeGameControllerListener(this);
      }

      currentGameController = gameController;
      currentGameController.addGameControllerListener(this);

      desktop.newGameStarted();
      desktop.openInterface();

      if (currentGameController.getCurrentPlayer() instanceof PlayerEngine) {
        this.computeNextMove();
      }
    } catch (NosResourcesException e) {
      logger.log(Level.SEVERE, "Resources exception", e);
    }

  }

  public GameController getCurrentGameController() {
    return currentGameController;
  }

  public void computeNextMove() {
    desktop.closeInterface();
    try {
      currentGameController.computeNextMove(true);
    } catch (GameControllerException e) {
      logger.log(Level.SEVERE, "Unexpected error:", e);
    }
  }

  public void engineMoveNow() {
    currentGameController.engineMoveNow();
  }

  private void announceGameEnd() {
    try {
      GameState gameState = currentGameController.getCurrentGameState();

      if (gameState.whiteMates()) {
        logger.info("Game finished. White mates!");
        desktop.announceGameEnd("1-0 {White mates}");
      } else if (gameState.blackMates()) {
        logger.info("Game finished. Black mates!");
        desktop.announceGameEnd("0-1 {Black mates}");
      } else if (gameState.remisByMaterial()) {
        logger.info("Game finished. Remis by material!");
        desktop.announceGameEnd("1/2-1/2 {Draw by material}");
      } else if (gameState.remisByFifty()) {
        logger.info("Game finished. Remis by 50 moves rule!");
        desktop.announceGameEnd("1/2-1/2 {Draw by repetition 50}");
      } else if (gameState.remisByThree()) {
        logger.info("Game finished. Remis by 3 times board repetition!");
        desktop.announceGameEnd("1/2-1/2 {Draw by position repetition 3}");
      } else if (gameState.remisByStaleMate()) {
        logger.info("Game finished. Remis by stale mate!");
        desktop.announceGameEnd("1/2-1/2 {Draw by Stale Mate}");
      }

    } catch (GameControllerException e) {
      logger.log(Level.SEVERE, "Unexpected error", e);
    }
  }

  public boolean isPromotionMove(MoveInput moveInput) {
    try {
      return currentGameController.isPromotionMove(moveInput);
    } catch (GameControllerException err) {
      logger.log(Level.SEVERE, "Unexpected error", err);
      return false;
    }
  }

  public void enterMove(MoveInput move) {
    try {
      currentGameController.enterNextMove(move);

      // Check for game end
      if (currentGameController.getCurrentGameState().gameOpen()) {
        if (currentGameController.getCurrentPlayer() instanceof PlayerEngine) {
          desktop.closeInterface();
          desktop.enteredMoveAcceptedEngineTurn();
          desktop.engineIsComputing();
          currentGameController.computeNextMove(true);
        } else {
          desktop.enteredMoveAcceptedHumanTurn();
          desktop.openInterface();
        }
      } else {
        desktop.enteredMoveAcceptedHumanTurn();
        announceGameEnd();
      }

    } catch (GameControllerException e) {
      desktop.enteredMoveRejected();
    }
  }

  public void takeBackMove() throws GameException {
    currentGameController.takeBackMove();
    desktop.moveTakenBack();
    desktop.openInterface();
  }

  public void loadGame(File pgnFile) {
    try {
      GameController gc = PGN.importFirstPGN(pgnFile);
      currentGameController = gc;
      currentGameController.addGameControllerListener(this);
      desktop.gameLoaded();
      desktop.openInterface();

    } catch (GameControllerException e) {
      desktop.showMessage("Load failed!");
    }

  }

  public void saveGame(File pgnFile) {
    try {
      PGN.exportPGN(currentGameController, pgnFile);
      desktop.showMessage("Game saved.");
    } catch (GameControllerException e) {
      desktop.showMessage("Save failed!");
    }

  }

  public String getPGN() {
    try {
      return currentGameController.getPGN();
    } catch (GameControllerException e) {
      logger.log(Level.SEVERE, "Unexpected error", e);
      return "";
    }

  }

  public Move getLastMove() {
    return currentGameController.getLastMove();
  }

  public Board getCurrentBoard() {
    return currentGameController.getCurrentBoard();
  }

  public String getGameHeader() {
    try {
      return currentGameController.getPlayerWhite().getPlayerName() + "   -   " + currentGameController.getPlayerBlack().getPlayerName();
    } catch (PlayerException e) {
      logger.log(Level.SEVERE, "Unexpected error", e);
      return "";

    }

  }

  public void computeNextLegalMovesFinished(GameControllerResult gameControllerResult) {

    desktop.notifyComputeNextMoveDepthFinished(gameControllerResult);
  }

  public void computeNextMoveFinished(GameControllerResult gameControllerResult) {
    try {
      desktop.notifyComputeNextMoveFinished(gameControllerResult);

      // Check for game end
      if (currentGameController.getCurrentGameState().gameOpen()) {
        if (currentGameController.getCurrentPlayer() instanceof PlayerEngine) {
          this.computeNextMove();
        } else {
          desktop.openInterface();
        }
      } else {
        desktop.openInterface();
        announceGameEnd();
      }

    } catch (GameControllerException err) {
      logger.log(Level.SEVERE, "Unexpected error", err);
    }

  }

  public void computeNextMoveDepthFinished(GameControllerResult engineResult) {
    desktop.notifyComputeNextMoveDepthFinished(engineResult);
  }

  public static void endApp() {
    logger.info("Shutdown Tiffanys.");
    System.exit(0);
  }

  public DesktopSettings getSettings() {
    return settings;
  }

  public void setSettings(DesktopSettings settings) {
    this.settings = settings;
  }

  public Desktop getDesktop() {
    return desktop;
  }

  /**
   * Show error dialog box, if possible
   * 
   * @param msg
   *          the msg
   * @param err
   *          the err
   */
  public void showError(String msg, Throwable err) {
    try {

      Frame currentFrame = null;
      Frame[] frames = Frame.getFrames();
      for (int i = 0; i < frames.length; i++) {
        if (frames[i].isVisible()) {
          currentFrame = frames[i];
        }
      }

      StringWriter stringWriter = new StringWriter();
      PrintWriter w = new PrintWriter(stringWriter);
      err.printStackTrace(w);
      String message = "Please enter a bug report at http://sourceforge.net/tracker/?group_id=35984  \nwith a short description and the error stack trace.\n\n" + stringWriter.toString();
      JOptionPane.showMessageDialog(currentFrame, new JTextArea(message, 5, 10), "Uncaught Exception!", JOptionPane.ERROR_MESSAGE);

    } catch (Throwable ignored) {
      // ignore, as we're allready in trouble :-(
    }

  }

}
