package ch.nostromo.tiffanys.engines.tiffanys;

import java.util.Arrays;
import java.util.Random;

import ch.nostromo.tiffanys.engines.Engine;
import ch.nostromo.tiffanys.engines.EngineException;
import ch.nostromo.tiffanys.engines.EngineResult;
import ch.nostromo.tiffanys.engines.EngineSettings;
import ch.nostromo.tiffanys.engines.tiffanys.ai.TiffanysAiTools;
import ch.nostromo.tiffanys.engines.tiffanys.board.TiffanysBoard;
import ch.nostromo.tiffanys.engines.tiffanys.board.TiffanysMove;
import ch.nostromo.tiffanys.engines.tiffanys.board.pieces.TiffanysPiece;
import ch.nostromo.tiffanys.game.board.Board;
import ch.nostromo.tiffanys.game.game.Game;
import ch.nostromo.tiffanys.game.move.MoveInput;

public class TiffanysEngine extends Engine {

  private int stats_positions_evaluated = 0;

  private int stats_ab_calls = 0;

  private int stats_ab_cuts = 0;

  private long searchStarted = 0;

  private TiffanysMove[] currentMoveList;

  private int currentPlannedSearchDepth = 0;

  private static Random random = null;

  private boolean abortable = false;

  public TiffanysEngine() {
    TiffanysPiece.initialize();
    TiffanysAiTools.initialize();
  }

  private int convertColor(int colorToMove) {
    if (colorToMove == Game.COLOR_WHITE) {
      return Board.OWNER_WHITE;
    } else {
      return Board.OWNER_BLACK;
    }
  }

  @Override
  public EngineResult engineSelectNextMove(Board pBoard, int colorToMove) {
    initAi();

    // Convert Board
    TiffanysBoard tboard = new TiffanysBoard(pBoard);
    int[] mainBoard = tboard.getBoard();

    // Convert color
    int color = convertColor(colorToMove);

    // First level ... check for check & sort by best move !!
    currentMoveList = TiffanysBoard.generateLegalMovesList(mainBoard, color);
    currentPlannedSearchDepth = 0;
    while (running) {
      try {
        currentPlannedSearchDepth++;

        // Sort moves, if several moves are equally good, take a random
        // one .
        for (int i = 0; i < currentMoveList.length; i++) {
          int[] board = mainBoard.clone();
          TiffanysBoard.makeMove(board, currentMoveList[i], color);

          int score = -alphaBeta(currentPlannedSearchDepth - 1, -MAX_RANGE, MAX_RANGE, board, TiffanysBoard.invertColor(color),  currentMoveList[i].hitmove);

          currentMoveList[i].score = score;
          currentMoveList[i].scoreDepth = currentPlannedSearchDepth;
          
          

        }

      } catch (EngineException ignored) {
        // Ignore, as search is interrupted by time
      }

      Arrays.sort(currentMoveList);
      this.processDepthFinished(createBestCurrentResult());
      
      // Abort search if winning mate is found
      boolean winningMoveFound = false;
      for (int i = 0; i < currentMoveList.length; i++) {
        if (currentMoveList[i].score >= MAX_MATE_RANGE) {
          winningMoveFound = true;
        }
      }
      if (winningMoveFound) {
        break;
      }
      
      // Abort search if only mate results
      boolean goOn = false;
      for (int i = 0; i < currentMoveList.length; i++) {
        if (currentMoveList[i].score > -MAX_MATE_RANGE) {
          goOn = true;
        }
      }
      if (!goOn) {
        break;
      }
      
      

      // Abort if fixed depth is reached
      if (engineSettings.GAME_MODE == EngineSettings.GAME_MODE_FIXED_DEPTH && engineSettings.FIXED_DEPTH == currentPlannedSearchDepth) {
        break;
      }

      abortable = true;
    }

    return createBestCurrentResult();

  }

  private EngineResult createBestCurrentResult() {
    TiffanysMove[] tiffMoves = currentMoveList.clone();
    Arrays.sort(tiffMoves);
    
    TiffanysMove selectedTiffMove = selectBestMove(tiffMoves);

    MoveInput selectedMove = null;
    if (selectedTiffMove != null) {
      selectedMove = selectedTiffMove.convertToMoveInput();
    }

    MoveInput[] moves = new MoveInput[tiffMoves.length];

    for (int i = 0; i < moves.length; i++) {
      moves[i] = tiffMoves[i].convertToMoveInput();
    }

    return new EngineResult(selectedMove, moves, dumpStats());

  }

  @Override
  public String getEngineName() {
    return "Tiffanys Engine 3";
  }

  private void initAi() {
    if (random == null) {
      random = new Random();
    }

    stats_ab_calls = 0;
    stats_ab_cuts = 0;
    stats_positions_evaluated = 0;
    searchStarted = System.currentTimeMillis();
  }

  private String dumpStats() {

    double ms = System.currentTimeMillis() - searchStarted;
    if (ms == 0)
      ms = 1;

    int evSec = (int) ((stats_positions_evaluated) / ms * 1000);

    String result = "";

    result += "Search depth: " + this.currentPlannedSearchDepth + "\n";
    result += "Legal moves evaluated: " + this.currentMoveList.length + "\n";
    result += "Total time in sec: " + ms / 1000 + "\n";
 //   result += "Alpha/Beta calls: " + stats_ab_calls + "\n";
 //   result += "Alpha/Beta cuts: " + stats_ab_cuts + "\n";
    result += "Positions evaluated: " + stats_positions_evaluated + " (" + evSec + " P/S) \n";

    return result;

  }

  private int evaluatePosition(int[] board, int currentColor) {
    stats_positions_evaluated++;

    int whitePos = 0;
    int blackPos = 0;

    for (int i = 21; i <= 98; i++) {

      if ((Board.OWNER_WHITE & board[i]) != 0) {

        if ((Board.PIECE_PAWN & board[i]) != 0) {
          whitePos += TiffanysAiTools.PAWN_WHITE_VALUES[i];
        } else {
          whitePos += TiffanysAiTools.COMMON_BOARD_VALUES[i];
        }

      } else if ((Board.OWNER_BLACK & board[i]) != 0) {

        if ((Board.PIECE_PAWN & board[i]) != 0) {
          blackPos += TiffanysAiTools.PAWN_BLACK_VALUES[i];
        } else {
          blackPos += TiffanysAiTools.COMMON_BOARD_VALUES[i];
        }
      }

    }

    int result = 0;

    if (currentColor == Board.OWNER_WHITE) {
      result = board[TiffanysBoard.IDX_WHITE_MATERIAL] - board[TiffanysBoard.IDX_BLACK_MATERIAL];
      result += whitePos;
      result -= blackPos;
    } else {
      result = board[TiffanysBoard.IDX_BLACK_MATERIAL] - board[TiffanysBoard.IDX_WHITE_MATERIAL];
      result += blackPos;
      result -= whitePos;
    }

    return result;
  }

  private int alphaBeta(int depth, int alpha, int beta, int[] board, int colorToMove, boolean fromHitMove) throws EngineException {

    int currentRelativeSearchDepth = this.currentPlannedSearchDepth - depth;

    if (!running && abortable) {
      throw new EngineException("Abort search");
    }

    if ((depth == 0 && !fromHitMove) || (currentRelativeSearchDepth > currentPlannedSearchDepth + engineSettings.MAX_QUIESCENCE_DEPTH)) {
      return evaluatePosition(board, colorToMove);
    }

    stats_ab_calls++;

    TiffanysMove[] moves = TiffanysBoard.generateMoveList(board, colorToMove);

    int val = 0;
    boolean legalMovefound = false;
    boolean hitMoveFound = false;
    for (int i = 0; i < moves.length; i++) {

      if (moves[i] != null) {

        TiffanysBoard.makeMove(board, moves[i], colorToMove);

        boolean isCheck = TiffanysBoard.checkForCheck(board, colorToMove);

        if (!isCheck) {
          legalMovefound = true;

          if (depth > 0 || moves[i].hitmove) {

            hitMoveFound = true;

            val = -alphaBeta(depth - 1, -beta, -alpha, board, TiffanysBoard.invertColor(colorToMove), moves[i].hitmove);

            if (val >= beta) {
              stats_ab_cuts++;
              TiffanysBoard.takeBackMove(board, moves[i]);
              return beta;
            }

            if (val > alpha) {
              alpha = val;
            }

          }

        }

        TiffanysBoard.takeBackMove(board, moves[i]);

      }
    }

    if (depth > 0 && !legalMovefound) {
      if (TiffanysBoard.checkForCheck(board, colorToMove)) {
        stats_positions_evaluated++;
        return -MAX_RANGE + currentRelativeSearchDepth;
      } else {
        stats_positions_evaluated++;
        return 0;
      }
    } else if (depth <= 0 && !hitMoveFound && legalMovefound) {
      return evaluatePosition(board, colorToMove);
    } else if (depth <= 0 && !hitMoveFound && !legalMovefound) {
      stats_positions_evaluated++;
      return 0;
    }

    return alpha;
  }

  private TiffanysMove selectBestMove(TiffanysMove[] legalMoves) {

    int change = 0;
    int value = legalMoves[0].score;
    for (int i = 0; i < legalMoves.length; i++) {
      if (legalMoves[i].score != value) {
        change = i;
        break;
      }
    }

    if (change > 0) {
      change = random.nextInt(change);
    }

    return (legalMoves[change]);

  }

}